import jax.numpy as jnp
from jax import jit
import numpy as np


@jit
def compute_msd_direct(x: jnp.ndarray) -> jnp.ndarray:
    """Compute the Mean Squared Displacement (MSD) using direct method.

    Parameters
    ----------
    x : jnp.ndarray
        Positions of a single atom in cartesian coordinates, shape (n_frames, 3).
    Returns
    -------
    jnp.ndarray
        Mean Squared Displacement for each time step, shape (n_frames,).
    """

    N = x.shape[0]

    def msd_at_dt(dt):
        displacements = x[dt:] - x[:-dt]
        squared_displacements = jnp.sum(displacements**2, axis=1)
        return jnp.mean(squared_displacements)

    msd = jnp.zeros(N)

    # Avoid looping over dt=0 because displacement is zero
    msd = msd.at[0].set(0.0)
    msd_values = jnp.array([msd_at_dt(dt) for dt in range(1, N)])
    msd = msd.at[1:].set(msd_values)

    return msd


def autocorrelation_1d_jax(data, N, n_fft):
    # Pad the signal with zeros
    padded_data = jnp.zeros(2 * n_fft).at[:N].set(data)

    # FFT → multiply by conjugate → IFFT → normalize
    fft_data = jnp.fft.fft(padded_data)
    autocorr = jnp.fft.ifft(fft_data * jnp.conj(fft_data))[:N].real
    norm = N - jnp.arange(N)  # Normalization factor
    return autocorr / norm


autocorrelation_1d_jax_jit = jit(autocorrelation_1d_jax, static_argnames=("N", "n_fft"))


def fn1(rsq, SAB, SUMSQ, N):
    MSD_0 = SUMSQ - 2 * SAB[0] * N

    cs1 = jnp.cumsum(rsq)[:-1]

    rsq_tail_reversed = rsq[1:][::-1]
    cs2 = jnp.cumsum(rsq_tail_reversed)

    denom = N - 1 - jnp.arange(N - 1)
    MSD_rest = (SUMSQ - cs1 - cs2) / denom
    MSD_rest -= 2 * SAB[1:]

    MSD = jnp.concatenate([jnp.array([MSD_0]), MSD_rest])
    return MSD


def _compute_msd_fft(pos: np.ndarray, n_fft: int, N: int) -> jnp.ndarray:
    rsq = jnp.sum(pos**2, axis=1)

    SAB = autocorrelation_1d_jax(pos[:, 0], N, n_fft)
    for i in range(1, pos.shape[1]):
        SAB += autocorrelation_1d_jax(pos[:, i], N, n_fft)

    SUMSQ = 2 * np.sum(rsq)

    MSD = fn1(rsq, SAB, SUMSQ, N)

    return MSD


def compute_msd_fft(positions: np.ndarray) -> np.ndarray:
    pos = np.asarray(positions)
    N = len(pos)
    n_fft = 2 ** (jnp.ceil(jnp.log2(N))).astype(int).item()
    if pos.shape[0] == 0:
        return np.array([], dtype=pos.dtype)
    if pos.ndim == 1:
        pos = pos.reshape((-1, 1))

    _compute_msd_fft_jit = jit(_compute_msd_fft, static_argnames=("n_fft", "N"))

    return _compute_msd_fft_jit(pos.reshape((N, -1)), n_fft=n_fft, N=N)



# conftest.py
import pytest
from ase import units
from ase.lattice.cubic import FaceCenteredCubic
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.verlet import VelocityVerlet
from ase.calculators.emt import EMT
import znh5md

import massband


@pytest.fixture
def emt_md(tmp_path):
    """Generate a fake MD trajectory by rattling atoms on a grid."""

    # TODO: need better data, e.g. containing molecules

    size = 3

    atoms = FaceCenteredCubic(
        directions=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        symbol="Cu",
        size=(size, size, size),
        pbc=True,
    )

    atoms.calc = EMT()

    MaxwellBoltzmannDistribution(atoms, temperature_K=300)

    dyn = VelocityVerlet(atoms, 5 * units.fs)  # 5 fs time step.

    io = znh5md.IO(
        filename=tmp_path / "md.h5",
        variable_shape=False,
    )

    def save_atoms():
        io.append(atoms)

    dyn.attach(save_atoms, interval=10)
    dyn.run(100)
    return tmp_path / "md.h5"


def test_RadialDistributionFunction(emt_md):
    """Test the radial distribution function calculation."""
    rdf = massband.RadialDistributionFunction(
        file=emt_md,
        batch_size=10,
    )

    # Calculate RDF
    rdf.run()

    # # Check if the result is a numpy array
    # assert isinstance(rdf.result, np.ndarray)

    # # Check if the shape is correct
    # assert rdf.result.shape == (int(5.0 / 0.1) + 1,)

    # # Check if the values are non-negative
    # assert np.all(rdf.result >= 0)



import lazy_loader as lazy
import logging
import sys

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
# attach stdout logger
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)



from .diffusion import EinsteinSelfDiffusion, KinisiSelfDiffusion
from .rdf import RadialDistributionFunction

__all__ = ["RadialDistributionFunction", "EinsteinSelfDiffusion", "KinisiSelfDiffusion"]



import logging
import pickle
from pathlib import Path
from typing import Optional, Union
from dataclasses import dataclass

import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import zntrack
from ase import Atoms
from collections import defaultdict
from functools import partial
from jax import vmap
from massband.utils import unwrap_positions
from znh5md import IO
import rdkit2ase
from kinisi.analyze import DiffusionAnalyzer

log = logging.getLogger(__name__)


@dataclass
class DiffusionPlotData:
    structure: str
    dt: np.ndarray
    msd: np.ndarray
    msd_std: np.ndarray
    distribution: np.ndarray
    D_samples: np.ndarray
    D_n: float


class KinisiSelfDiffusion(zntrack.Node):
    file: Union[str, Path] = zntrack.deps_path()
    sampling_rate: int = zntrack.params()
    time_step: float = zntrack.params()
    structures: Optional[list[str]] = zntrack.params(None)
    start_dt: float = zntrack.params(50)  # in ps

    data_path: Path = zntrack.outs_path(zntrack.nwd / "diffusion_data")

    def get_data(self):
        io = IO(self.file, variable_shape=False, include=["position", "box"])
        frames: list[Atoms] = io[:]
        positions = jnp.stack([atoms.positions for atoms in frames])
        cells = jnp.stack([atoms.cell[:] for atoms in frames])
        inv_cells = jnp.linalg.inv(cells)

        return {
            "positions": positions,
            "cells": cells,
            "inv_cells": inv_cells,
            "frames": frames,
        }

    def run(self):
        data = self.get_data()
        unwrap_positions_vmap = vmap(
            partial(unwrap_positions, cells=data["cells"], inv_cells=data["inv_cells"]),
            in_axes=(1,),
        )
        positions_unwrapped = unwrap_positions_vmap(data["positions"])
        log.info(f"Unwrapped positions shape: {positions_unwrapped.shape}")

        substructures = defaultdict(list)
        if self.structures:
            log.info(f"Searching for substructures in {len(self.structures)} patterns")
            for structure in self.structures:
                indices = rdkit2ase.match_substructure(
                    data["frames"][0],
                    smiles=structure,
                    suggestions=self.structures,
                )
                if indices:
                    substructures[structure].extend(indices)
                    log.info(f"Found {len(indices)} matches for {structure}")

        self.data_path.mkdir(exist_ok=True)

        for structure, indices in substructures.items():
            flat_indices = [i for sublist in indices for i in sublist]
            sub_frames = [atoms[flat_indices] for atoms in data["frames"]]
            specie_indices = rdkit2ase.match_substructure(
                sub_frames[0],
                smiles=structure,
                suggestions=self.structures,
            )
            masses = sub_frames[0][specie_indices[0]].get_masses().tolist()

            diff = DiffusionAnalyzer.from_ase(
                sub_frames,
                parser_params={
                    "specie": None,
                    "time_step": self.time_step / 1000,
                    "step_skip": self.sampling_rate,
                    "specie_indices": specie_indices,
                    "masses": masses,
                    "progress": True,
                },
                uncertainty_params={"progress": True},
            )
            diff.diffusion(self.start_dt, {"progress": True})

            result = DiffusionPlotData(
                structure=structure,
                dt=np.asarray(diff.dt),
                msd=np.asarray(diff.msd),
                msd_std=np.asarray(diff.msd_std),
                distribution=np.asarray(diff.distribution),
                D_samples=np.asarray(diff.D.samples),
                D_n=float(diff.D.n),
            )

            with open(self.data_path / f"{structure}.pkl", "wb") as f:
                pickle.dump(result, f)

        self.plot()

    def plot(self):
        credible_intervals = [[16, 84], [2.5, 97.5], [0.15, 99.85]]
        alpha = [0.6, 0.4, 0.2]

        for pkl_path in self.data_path.glob("*.pkl"):
            with open(pkl_path, "rb") as f:
                data: DiffusionPlotData = pickle.load(f)

            # MSD with std
            fig, ax = plt.subplots()
            ax.errorbar(data.dt, data.msd, data.msd_std)
            ax.set_ylabel("MSD/Å$^2$")
            ax.set_xlabel(r"$\Delta t$/ps")
            ax.set_title(f"{data.structure} MSD with std")
            fig.savefig(self.data_path / f"{data.structure}_msd_std.png", dpi=300)
            plt.close(fig)

            # MSD with credible intervals
            fig, ax = plt.subplots()
            ax.plot(data.dt, data.msd, "k-")
            for i, ci in enumerate(credible_intervals):
                low, high = np.percentile(data.distribution, ci, axis=1)
                ax.fill_between(
                    data.dt, low, high, alpha=alpha[i], color="#0173B2", lw=0
                )
            # TODO: save start_dt in pickle as well?
            ax.axvline(self.start_dt, c="k", ls="--")
            ax.set_ylabel("MSD/Å$^2$")
            ax.set_xlabel(r"$\Delta t$/ps")
            ax.set_title(f"{data.structure} MSD credible intervals")
            fig.savefig(
                self.data_path / f"{data.structure}_credible_intervals.png", dpi=300
            )
            plt.close(fig)

            # Histogram of diffusion coefficients
            fig, ax = plt.subplots()
            ax.hist(data.D_samples, density=True, bins=50)
            ax.axvline(data.D_n, c="k")
            ax.set_xlabel("$D$/cm$^2$s$^{-1}$")
            ax.set_ylabel("$p(D)$/cm$^2$s$^{-1}$")
            ax.set_title(f"{data.structure} Diffusion Histogram")
            fig.savefig(self.data_path / f"{data.structure}_hist.png", dpi=300)
            plt.close(fig)



import jax.numpy as jnp
import numpy as np
import pytest

from massband.utils import unwrap_positions


@pytest.fixture()
def wrapped_trajectory():
    """Create a wrapped trajectory for testing."""
    # Create a simple trajectory with periodic boundary conditions
    cell = np.eye(3)  # 1 Angstrom cubic cell
    pos1 = [0.5, 0.5, 0.5]  # Start at the center of the cell
    pos2 = [0.9, 0.9, 0.9]  # Move close to the edge of the cell
    pos3 = [0.1, 0.1, 0.1]  # Assume movement across the boundary

    # particle moving twice across the boundary
    return (
        np.array([pos1, pos2, pos3, pos1, pos2, pos3, pos1, pos2, pos3]),
        np.array([cell] * 9),
        np.array([np.linalg.inv(cell)] * 9),
    )


def test_unwrap_positions(wrapped_trajectory):
    positions, cell, inv_cell = wrapped_trajectory
    # raise ValueError(f"Unwrap positions test: {positions.shape}, {cell.shape}, {inv_cell.shape}")

    # Unwrap the positions
    unwrapped_positions = unwrap_positions(positions, cell, inv_cell)

    # Check if the unwrapped positions are correct
    expected_unwrapped = np.array(
        [
            [0.5, 0.5, 0.5],
            [0.9, 0.9, 0.9],
            [1.1, 1.1, 1.1],
            [1.5, 1.5, 1.5],
            [1.9, 1.9, 1.9],
            [2.1, 2.1, 2.1],
            [2.5, 2.5, 2.5],
            [2.9, 2.9, 2.9],
            [3.1, 3.1, 3.1],
        ]
    )

    assert jnp.allclose(unwrapped_positions, expected_unwrapped)



from collections import defaultdict
import logging
from typing import Dict, List, Tuple, Literal

import jax.numpy as jnp
import matplotlib.pyplot as plt
import pint
import znh5md
import zntrack
from ase.data import chemical_symbols
from jax import jit, vmap
from tqdm import tqdm
from massband.utils import unwrap_positions
import rdkit2ase
from massband.diffusion.utils import compute_msd_direct, compute_msd_fft
import jax

# Enable 64-bit precision in JAX for FFT accuracy
jax.config.update("jax_enable_x64", True)

ureg = pint.UnitRegistry()
log = logging.getLogger(__name__)


class EinsteinSelfDiffusion(zntrack.Node):
    """Compute self-diffusion coefficients using Einstein relation from MD trajectories."""

    file: str = zntrack.deps_path()
    sampling_rate: int = zntrack.params()
    timestep: float = zntrack.params()
    batch_size: int = zntrack.params(64)
    fit_window: Tuple[float, float] = zntrack.params((0.2, 0.8))
    method: Literal["direct", "fft"] = zntrack.params("fft")
    structures: list[str] | None = zntrack.params(None)
    # TODO: allow not smiles but just the sum formula, e.g. "C6H12O6" for glucose

    def get_cells(self) -> Tuple[jnp.ndarray, jnp.ndarray]:
        io = znh5md.IO(
            self.file, variable_shape=False, include=["position", "box"], mask=[]
        )
        cells = jnp.stack([atoms.cell[:] for atoms in io[:]])
        inv_cells = jnp.linalg.inv(cells)
        return cells, inv_cells

    def get_atomic_numbers(self) -> List[int]:
        """Get atomic numbers from the trajectory."""
        io = znh5md.IO(self.file, variable_shape=False, include=["position"])
        return io[0].get_atomic_numbers().tolist()

    def get_masses(self) -> List[float]:
        """Get atomic masses from the trajectory."""
        io = znh5md.IO(self.file, variable_shape=False, include=["position"])
        return io[0].get_masses().tolist()

    def get_positions(self, index: slice) -> jnp.ndarray:
        """Load unwrapped positions for a slice of atom indices."""
        # TODO: for COM diffusion, we should process this here ?

        log.info(f"Loading positions for index {index}")
        io = znh5md.IO(
            self.file, variable_shape=False, include=["position"], mask=index
        )
        pos = jnp.stack([atoms.positions for atoms in io[:]])
        log.info(f"Loaded positions shape: {pos.shape}")
        return pos  # shape: (n_frames, n_atoms_in_batch, 3)

    def postprocess_positions(
        self,
        pos: jnp.ndarray,
        masses,
        atomic_numbers,
        substructures,
        atom_slice: slice,
        max_cache_size: int = 10000,
    ):
        # pos: (n_frames, n_atoms_in_batch, 3)
        # masses: (n_total_atoms,)
        # atomic_numbers: (n_total_atoms,)
        # substructures: dict[substructure_name -> list[tuple[atom_indices]]]
        # atom_slice: slice (start, stop) for current batch of atoms

        values = []
        Zs = []

        try:
            cache = getattr(self, "data_cache")
        except AttributeError:
            cache = {}

        # Track all molecular atom indices
        molecular_indices = set()
        mol_index_map = {}  # atom idx -> list of (substructure_name, mol_indices)

        for sub_name, mols in substructures.items():
            for mol_indices in mols:
                for idx in mol_indices:
                    mol_index_map.setdefault(idx, []).append(
                        (sub_name, tuple(mol_indices))
                    )
                    molecular_indices.add(idx)

        # Cache positions of molecular atoms, append standalone atoms directly
        for i in range(atom_slice.start, atom_slice.stop):
            rel_i = i - atom_slice.start
            pos_i = pos[:, rel_i, :]  # (n_frames, 3)

            if i in molecular_indices:
                cache[i] = pos_i
            else:
                Z = atomic_numbers[i]
                values.append(pos_i)
                Zs.append(Z)

        # Enforce cache size
        if len(cache) > max_cache_size:
            raise RuntimeError(f"Cache size exceeded max limit of {max_cache_size}")

        # Track used molecules so we can remove them from cache
        used_mol_keys = set()

        for sub_name, mols in substructures.items():
            for mol_indices in mols:
                if all(idx in cache for idx in mol_indices):
                    # All atoms in molecule are available, compute COM
                    pos_stack = jnp.stack(
                        [cache[idx] * masses[idx] for idx in mol_indices], axis=0
                    )  # (n_atoms, n_frames, 3)
                    total_mass = jnp.sum(
                        jnp.array([masses[idx] for idx in mol_indices])
                    )
                    com = jnp.sum(pos_stack, axis=0) / total_mass  # (n_frames, 3)

                    values.append(com)
                    Zs.append(sub_name)

                    used_mol_keys.add(tuple(mol_indices))

        # Clear used entries from cache
        for mol_indices in used_mol_keys:
            for idx in mol_indices:
                cache.pop(idx, None)
        print(
            f"Used {len(used_mol_keys)} molecular indices, remaining cache size: {len(cache)}"
        )

        self.data_cache = cache

        values = jnp.stack(values, axis=1)  # shape: (n_frames, n_entities, 3)
        return Zs, values

    def compute_diffusion_coefficients(
        self, msds: Dict[int, List[jnp.ndarray]], timestep_fs: float
    ) -> Dict:
        """Calculate diffusion coefficients from MSD data.

        Args:
            msds: Dictionary mapping atomic numbers to lists of MSD arrays.
            timestep_fs: MD timestep in femtoseconds.

        Returns:
            Dictionary containing full data for plotting and analysis.
        """
        results = {}
        timestep_ps = timestep_fs / 1000  # fs -> ps

        for Z, msd_array in msds.items():
            msd_avg = jnp.mean(jnp.stack(msd_array), axis=0)
            time_ps = jnp.arange(msd_avg.shape[0]) * timestep_ps

            # Fit to linear region
            start_idx = int(len(time_ps) * self.fit_window[0])
            end_idx = int(len(time_ps) * self.fit_window[1])

            fit_time = time_ps[start_idx:end_idx]
            fit_msd = msd_avg[start_idx:end_idx]

            # Linear least squares fit
            A = jnp.vstack([fit_time, jnp.ones_like(fit_time)]).T
            slope, intercept = jnp.linalg.lstsq(A, fit_msd, rcond=None)[0]
            D_fit = slope / 6
            fit_line = slope * time_ps + intercept
            try:
                symbol = chemical_symbols[Z]
            except TypeError:
                symbol = Z

            results[Z] = {
                "Z": Z,
                "symbol": symbol,
                "diffusion_coefficient": D_fit,
                "time_ps": time_ps,
                "msd": msd_avg,
                "fit_line": fit_line,
                "fit_window": (start_idx, end_idx),
                "fit_msd": fit_msd,
                "fit_time": fit_time,
                "slope": slope,
                "intercept": intercept,
            }

            log.info(f"Z={Z} ({symbol}): D = {D_fit:.4f} Å²/ps")

        return results

    def plot_results(self, results: Dict, filename: str = "msd_species.png"):
        """Plot MSD curves and diffusion coefficient fits.

        Args:
            results: Dictionary containing precomputed MSD and fit data
            filename: Output filename for the plot
        """
        n_species = len(results)
        n_cols = min(3, n_species)
        n_rows = (n_species + n_cols - 1) // n_cols
        fig, axes = plt.subplots(n_rows, n_cols, figsize=(5 * n_cols, 4 * n_rows))

        if n_rows == 1 and n_cols == 1:
            axes = [axes]
        else:
            axes = axes.flatten()

        for ax, (Z, data) in zip(axes, results.items()):
            time_ps = data["time_ps"]
            msd = data["msd"]
            fit_line = data["fit_line"]
            start_idx, end_idx = data["fit_window"]

            ax.plot(time_ps, msd, label="MSD")
            ax.plot(
                time_ps,
                fit_line,
                "--",
                label=f"Fit: D = {data['diffusion_coefficient']:.4f} Å²/ps",
                color="tab:red",
            )
            ax.axvspan(
                time_ps[start_idx],
                time_ps[end_idx - 1],
                color="gray",
                alpha=0.2,
                label="Fit Window",
            )

            ax.set_title(f"{data['symbol']} (Z={Z})")
            ax.set_xlabel("Time (ps)")
            ax.set_ylabel("MSD (Å²)")
            ax.legend()
            ax.grid(True)

        # Hide unused subplots
        for ax in axes[n_species:]:
            ax.set_visible(False)

        plt.tight_layout()
        plt.savefig(filename, dpi=300, bbox_inches="tight")
        plt.close()

    def run(self):
        """Main computation workflow."""
        log.info("Collecting cell vectors and inverse cell vectors")
        cells, inv_cells = self.get_cells()

        substructures = defaultdict(list)
        # dict[str, list[tuple[int, ...]]]
        if self.structures is not None:
            io = znh5md.IO(self.file, variable_shape=False)
            atoms = io[0]
            log.info(f"Searching for substructures in {len(self.structures)} patterns")
            for structure in self.structures:
                indices = rdkit2ase.match_substructure(
                    atoms,
                    smiles=structure,
                    suggestions=self.structures,
                )
                if len(indices) > 0:
                    substructures[structure].extend(indices)

                log.info(
                    f"Found {len(indices)} matches for substructure {structure} in the dataset."
                )
            # TODO: in this case, we don't want to compute the of each atom, but rather the center of mass of the molecule, given by the indices
            # - get the mass of each atom in the molecule from the initial structure using ase
            # - assume the indices are the same for all frames, iterate the dataset in the given batch size, for all indices not in the molecule, compute the MSD as usual
            # for all the others, store the positions in a dict[index] = positions, and as soon as we have all the positions for any molecule, compute the MSD for that molecule
            # using the COM and then remove the indices / positions from the dataset
            # have a max size of the data cache and raise an error, also allow direct indexing as an alternative.

        log.info("Collecting atomic masses and numbers")
        masses = self.get_masses()
        atomic_numbers = self.get_atomic_numbers()
        # use iddentifier and not aotmic numbers so one can also use com
        timestep_fs = (self.timestep * ureg.fs * self.sampling_rate).magnitude
        msds = defaultdict(list)

        n_atoms = len(atomic_numbers)
        log.info(f"Starting MSD calculation for {n_atoms} atoms")

        if self.method == "direct":

            @jit
            def msd_fn(x, cell, inv_cell):
                # x_unwrapped = unwrap_positions(x, cell, inv_cell)
                return compute_msd_direct(x)

            log.info("Using direct MSD computation method")
        elif self.method == "fft":

            def msd_fn(x, cell, inv_cell):
                # x_unwrapped = unwrap_positions(x, cell, inv_cell)
                return compute_msd_fft(x)

            log.info("Using FFT-based MSD computation method")
        else:
            raise ValueError(f"Unknown method: {self.method}. Use 'direct' or 'fft'.")

        # Process atoms in batches
        for start in tqdm(range(0, n_atoms, self.batch_size), desc="Processing atoms"):
            end = min(start + self.batch_size, n_atoms)
            atom_slice = slice(start, end)
            Z_batch = atomic_numbers[start:end]

            pos = self.get_positions(atom_slice)  # shape: (n_frames, batch_size, 3)
            # postprocess positions for substructures
            # TODO: need to unwrap here!!
            # pos = unwrap_positions(pos, cells, inv_cells)
            # TODO: fix unnecessary multiple transposes
            pos = jnp.transpose(pos, (1, 0, 2))
            pos = vmap(lambda x: unwrap_positions(x, cells, inv_cells))(pos)
            pos = jnp.transpose(pos, (1, 0, 2))
            Z_batch, pos = self.postprocess_positions(
                pos, masses, atomic_numbers, substructures, atom_slice
            )
            # each step print the shape of the cached pos
            if self.method == "direct":
                pos = jnp.transpose(pos, (1, 0, 2))
                # TODO, specify vmap axis instead of transposing
                results = vmap(lambda x: msd_fn(x, cells, inv_cells))(pos)
            else:
                # TODO: vmap this stuff
                results = []
                for atom_index in range(pos.shape[1]):
                    pos_i = pos[:, atom_index, :]
                    msd_i = msd_fn(pos_i, cells, inv_cells)
                    results.append(msd_i)
            for msd, Z in zip(results, Z_batch):
                msds[Z].append(msd)

        results = self.compute_diffusion_coefficients(msds, timestep_fs)
        self.results = results
        self.plot_results(results)



import itertools
from collections import defaultdict

import ase
import jax.numpy as jnp
import matplotlib.pyplot as plt
import znh5md
import zntrack
from ase.data import chemical_symbols
from jax import vmap
from laufband import Laufband


def compute_mic_distances(positions: jnp.ndarray, cells: jnp.ndarray) -> jnp.ndarray:
    """
    Computes the pairwise distance matrix for each frame, accounting for periodic boundary conditions
    using the minimum image convention (MIC).

    Parameters:
    - positions: (N, i, 3) array of atomic positions
    - cells: (N, 3, 3) array of unit cell matrices per frame

    Returns:
    - distances: (N, i, i) array of pairwise distances per frame
    """

    def mic_frame(pos: jnp.ndarray, cell: jnp.ndarray):
        """
        MIC-based distance computation for a single frame.

        pos: (i, 3)
        cell: (3, 3)
        returns: (i, i) distance matrix
        """
        inv_cell = jnp.linalg.inv(cell.T)  # (3, 3)
        frac = pos @ inv_cell  # (i, 3) fractional
        delta_frac = frac[:, None, :] - frac[None, :, :]  # (i, i, 3)
        delta_frac = delta_frac - jnp.round(delta_frac)  # wrap to [-0.5, 0.5)
        delta_cart = delta_frac @ cell.T  # back to Cartesian
        dists = jnp.linalg.norm(delta_cart, axis=-1)  # (i, i)
        return dists

    # Vectorize over frames (N)
    return vmap(mic_frame)(positions, cells)


def plot_rdf(rdfs: defaultdict):
    # find best number of subplots
    n_rdfs = len(rdfs)
    n_cols = 3
    n_rows = (n_rdfs + n_cols - 1) // n_cols  # Ceiling division
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(15, 5 * n_rows))
    axes = axes.flatten()

    for ax, ((a, b), g_r_list) in zip(axes, rdfs.items()):
        g_r_array = jnp.stack(g_r_list)
        g_r_mean = jnp.mean(g_r_array, axis=0)
        r = 0.5 * (jnp.arange(len(g_r_mean)) + 0.5) * 0.1
        ax.plot(r, g_r_mean, label=f"{chemical_symbols[a]}-{chemical_symbols[b]}")
        ax.set_xlabel("Distance r (Å)")
        ax.set_ylabel("g(r)")
        ax.set_title(
            f"Radial Distribution Function: {chemical_symbols[a]}-{chemical_symbols[b]}"
        )
        ax.legend()
        ax.grid(True)
    fig.tight_layout()
    # Save the figure
    fig.savefig("rdf_plot.png")
    plt.close(fig)  # Close the figure to free memory


class RadialDistributionFunction(zntrack.Node):
    """
    Class to represent a radial distribution function (RDF) in a molecular dynamics simulation.
    """

    file: str = zntrack.deps_path()
    batch_size: int = zntrack.params()  # You can set a default or make it configurable
    bin_width: float = zntrack.params(0.05)  # Width of the bins for RDF

    def run(self):
        io = znh5md.IO(self.file, variable_shape=False, include=["position", "box"])
        size = len(io)
        batch_start_index = list(range(0, size, self.batch_size))
        worker = Laufband(batch_start_index, cleanup=True)

        # Collect RDFs for each species pair over all batches
        rdfs_all = defaultdict(list)

        for start in worker:
            end = min(start + self.batch_size, size)
            batch = io[start:end]
            rdf_batch = self.compute_rdf(batch)
            for pair, (r, g_r) in rdf_batch.items():
                rdfs_all[pair].append(g_r)

        # Plot the results
        plot_rdf(rdfs_all)
        # TODO: use rdkit2ase to map substructures

    def compute_rdf(
        self, batch: list[ase.Atoms]
    ) -> dict[tuple[int, int], tuple[jnp.ndarray, jnp.ndarray]]:
        # Convert ASE objects to JAX arrays
        positions = jnp.stack(
            [jnp.array(atoms.positions) for atoms in batch]
        )  # (N, i, 3)
        cells = jnp.stack([jnp.array(atoms.cell[:]) for atoms in batch])  # (N, 3, 3)
        atomic_numbers = jnp.array(batch[0].get_atomic_numbers())  # (i,)

        N_frames, N_atoms, _ = positions.shape
        distances = compute_mic_distances(positions, cells)  # (N, i, i)

        # Only take upper triangle (i < j)
        i, j = jnp.triu_indices(N_atoms, k=1)  # (n_pairs,)
        species_i = atomic_numbers[i]  # (n_pairs,)
        species_j = atomic_numbers[j]  # (n_pairs,)

        pairwise_dists = distances[:, i, j].reshape(-1)  # (N_frames * n_pairs,)

        # Define bins
        r_max = 10.0
        bin_edges = jnp.arange(0, r_max + self.bin_width, self.bin_width)
        bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])
        shell_volumes = (4 / 3) * jnp.pi * (bin_edges[1:] ** 3 - bin_edges[:-1] ** 3)

        # Volume per frame (averaged)
        volume = jnp.mean(jnp.linalg.det(cells))

        # Unique atomic numbers
        unique_species = set(atomic_numbers.tolist())

        rdf_by_species = {}

        for a, b in itertools.combinations_with_replacement(sorted(unique_species), 2):
            # Mask for matching pairs (upper triangle only)
            frame_mask = ((species_i == a) & (species_j == b)) | (
                (species_i == b) & (species_j == a)
            )  # shape: (n_pairs,)
            frame_mask = jnp.asarray(frame_mask)

            # Repeat this mask for each frame
            full_mask = jnp.tile(frame_mask, N_frames)  # shape: (N_frames * n_pairs,)
            selected_dists = pairwise_dists[full_mask]

            if selected_dists.size == 0:
                continue  # skip empty

            # Histogram
            hist = jnp.histogram(selected_dists, bins=bin_edges)[0]

            # Estimate number density of this pair type
            n_pair_per_frame = frame_mask.sum()
            number_density = (n_pair_per_frame * N_frames) / volume

            ideal_counts = number_density * shell_volumes
            g_r = hist / ideal_counts

            rdf_by_species[(int(a), int(b))] = (bin_centers, g_r)

        return rdf_by_species



import jax.numpy as jnp
from jax import jit


@jit
def unwrap_positions(
    pos: jnp.ndarray, cells: jnp.ndarray, inv_cells: jnp.ndarray
) -> jnp.ndarray:
    """Unwrap positions using the inverse cell matrix.

    Parameters
    ----------
    pos : jnp.ndarray
        Positions of atoms in cartesian coordinates, shape (n_frames, 3).
    cells : jnp.ndarray
        Cell vectors for each frame, shape (n_frames, 3, 3).
    inv_cells : jnp.ndarray
        Inverse cell vectors for each frame, shape (n_frames, 3, 3).

    Returns
    -------
    jnp.ndarray
        Unwrapped positions in cartesian coordinates, shape (n_frames, 3).
    """
    # Convert to fractional coordinates per frame
    frac = jnp.einsum("nij,nj->ni", inv_cells, pos)

    # Compute wrapped displacements
    delta_frac = jnp.diff(frac, axis=0)
    delta_frac -= jnp.round(delta_frac)

    # Reconstruct unwrapped fractional coords
    frac_unwrapped = jnp.concatenate(
        [frac[:1], frac[:1] + jnp.cumsum(delta_frac, axis=0)], axis=0
    )

    # Back to cartesian coordinates
    pos_unwrapped = jnp.einsum("nij,nj->ni", cells, frac_unwrapped)
    return pos_unwrapped



import pytest
import numpy as np
import pint
from ase import Atoms
import massband
from massband.diffusion.utils import compute_msd_direct, compute_msd_fft
import numpy.testing as npt
import tidynamics
import znh5md
from pathlib import Path

ureg = pint.UnitRegistry()


@pytest.fixture
def diffusive_positions() -> np.ndarray:
    # Parameters
    D = 1.0 * ureg.angstrom**2 / ureg.picosecond
    steps = 100
    dt = 0.5 * ureg.fs
    particles = 20

    rng = np.random.default_rng(42)  # For reproducibility

    # Convert quantities to atomic units
    D_val = D.to("angstrom**2/fs").magnitude
    dt_val = dt.to("fs").magnitude

    # Initialize system
    frames = []
    pos = np.zeros((particles, 3))  # Start all particles at origin

    for _ in range(steps):
        # Random displacement with variance 2DΔt in each dimension
        displacement = rng.normal(
            loc=0.0,  # Mean of 0
            scale=np.sqrt(2 * D_val * dt_val),
            size=(particles, 3),
        )
        pos += displacement
        frames.append(pos.copy())

    return np.array(frames, dtype=np.float32), dt_val


@pytest.fixture
def diffusive_md(tmp_path, diffusive_positions) -> Path:
    """Create a trajectory with a fixed diffusion coefficient.

    Returns:
        List[ase.Atoms]: A trajectory where particles follow Brownian motion
        with the specified diffusion coefficient.
    """
    positions, dt_val = diffusive_positions
    particles = positions.shape[1]
    cell = np.eye(3) * 10.0
    dt_val = 0.5

    frames = []
    # Generate Brownian motion trajectory
    for position in positions:
        # Random displacement with variance 2DΔt in each dimension

        # Create ASE Atoms object for this step
        atoms = Atoms(
            f"H{particles}", positions=position, cell=cell, pbc=True
        )  # Add cell and pbc
        frames.append(atoms)

    io = znh5md.IO(
        filename=tmp_path / "diffusive_trajectory.h5",
        timestep=dt_val,
        variable_shape=False,
    )
    io.extend(frames)

    return tmp_path / "diffusive_trajectory.h5"


def test_EinsteinSelfDiffusion(diffusive_md):
    """Test the Einstein self-diffusion coefficient calculation."""
    diff = massband.EinsteinSelfDiffusion(
        file=diffusive_md,
        sampling_rate=1,
        timestep=0.5,  # fs
        batch_size=1,
        fit_window=(0.05, 0.95),  # we don't have a ballistic regime
    )

    diff.run()
    # Now we expect results to be calculated, not None
    assert diff.results is not None
    assert diff.results[1]["diffusion_coefficient"] == pytest.approx(1.0, rel=0.1)


def test_compute_msd_direct(diffusive_positions):
    """Test the direct computation of Mean Squared Displacement."""
    positions, _ = diffusive_positions
    msd_1 = compute_msd_direct(positions[:, 0])
    msd_2 = tidynamics.msd(positions[:, 0])
    msd_3 = compute_msd_fft(positions[:, 0])
    # Check that the two methods yield similar results, we ignore the first 10 and last 10 points to avoid edge effects of the fft
    npt.assert_allclose(msd_1[10:-10], msd_2[10:-10], rtol=5e-4)
    npt.assert_allclose(msd_1[10:-10], msd_3[10:-10], rtol=5e-4)
